#This script contains a function for predicting from DetHetGP

DetHetGPPredict <- function(x, y, xdet, ydet, parameterestimates, xpred, M = 1000) {
  #This function predicts from DetHetGP
  
  #needs input points "x", and outputs "y" from the stochastic simulator
  #as well as inputs "xdet" and outputs "ydet" from the deterministic simulator
  #And points to predict for xpred
  #Also needs a list "parameterestimates" (output from DetHetGPFit)
  #The final optional input M is the number of posterior draws to then take.
  
  #The output is a list of 5 sets of posterior predictive samples
  #The first object is the main output, and is the predictions from DetHetGP at the prediction points xpred, for most purposes this is all that is needed
  
  #The second is the predictions of the intrinsic variance at the prediction points
  #The third is the predictions of DetGP at the input points
  #The fourth is the predictions of DetGP at the prediction points
  #The fifth is the predictions of HetGP at the prediction points
  
  
  #this function works by sampling values from the variance GP, sampling values from DetGP, and then sampling a value of HetGP
  #Adding a value for DetGP and HetGP together obtains a sample for y.
  
  
  #This code could be substantially sped up, by taking the predictions for the variance GP, and the DetGP at the stochastic input point, fixed at their expectation.
  #this then makes the DetHetGP predictions analytically known, which would no longer require us to sample values of the variance GP, HetGP and DetGP
  #One could calculate the mean and variance of DetHetGP analytically.
  
  #In the paper this is not done, and so it is also not done in this code to provide consistency.

  library(MASS)
  library(Matrix)
  library(rstan)
  library(rdist)

#we want predictions for new points, but to do this we need a few functions

  #we need the squared exponential covariance function (these *could* be customised, but would require a change in Stan as well, and probably would be parametrised differently, i.e. no length scale)
  K12 <-  function(X1, X2, l) { 
    #Calculates squared exponential covariance matrix for given inputs
    n1 = length(X1[,1]) #number of input points
    n2 = length(X2[,1]) #number of validation points
    d = length(X1[1,])  #number of variables
    D = as.matrix(cdist(scale(X1,center=FALSE,scale=l),scale(X2,center=FALSE,scale=l))) 
    return(exp(-(D^2)))
  }

  
  #H matrices (for mean functions) (these can be customised, but changes need to replicated in the fit function)
  ##HetGP
  H <- function(x){
    #creates our "H" matrix for HetGP
    cbind(1,x) #linear mean function
  }
  Hmatrix <- H(x)               #get H matrix
  dimenH <- length(Hmatrix[1,]) #How many dimensions is it?
  
  ##variance GP
  varH <- function(x){
    #creates our "H" matrix for the variance GP
    cbind(1,x) #linear mean function
  }
  Hvar <- varH(x)              #get H matrix
  dimenHvar <- length(Hvar[1,]) #How many dimensions is it?
  
  ##DetGP
  detH <- function(x){
    #creates our "H" matrix for DetGP
    cbind(1,x) #linear mean function
  }
  Hdet <- detH(xdet)              #get H matrix for the deterministic points
  Hdet2 <- detH(x)                #get Hmatrix for the stochastic points
  dimenHdet <- length(Hdet[1,])   #How many dimensions are they?

  
  #Then given a H matrix, obtain the mean function values
  meanfunc <-  function(H, beta) {
    #Calculates (prior) mean function for a given GP
    H%*%as.vector(beta)
  }

  #The next set of functions are the respective variance GP, DetGP, and HetGP GP predictions. 
  #These are all effectively standard GP predictive equations with slight modifications
  #(HetGP also has input dependent noise (coming from the variance GP), and its prior mean function = meanfunc + DetGP)
  
  #######
  #Variance Prediction
  #######
  deltaGP <- function(inputx, inputy, predx, alpha, beta, l, nugget, sigma) {
    #generates a gaussian process for given inputs, outputs, prediction points, and parameter estimates
    #This GP will be for the variances (which makes it a standard GP prediction + additional estimated nugget)
  
    #We need A = K(X, X); Kyx = K(X*, X); Kyy = K(X*, X*); Kxy = K(X, X*)
    
    A <- (alpha^2) * K12(inputx, inputx, l) + diag(rep(nugget, length(inputx[,1])), nrow = length(inputx[,1])) + diag(rep(sigma, length(inputx[,1])), nrow = length(inputx[,1]))#plus nugget
    Ainv <- chol2inv(chol(A)) #get inverse
  
    Kyx <- (alpha^2) * K12(predx, inputx, l)
    Kxy <- t(Kyx)
    Kyy <- (alpha^2) * K12(predx, predx, l) + diag(rep(nugget, length(predx[,1])), nrow = length(predx[,1])) + diag(rep(sigma, length(predx[,1])), nrow = length(predx[,1]))#plus nugget 
  
    #Now get posterior mean 'function'
    #obtain prior mean functions for input and prediction points
    Hpred <- varH(predx)
    Hinput <- varH(inputx)
    postmean <- meanfunc(Hpred, beta) +Kyx%*%Ainv%*%(inputy - meanfunc(Hinput, beta)) #calculate posterior mean
  
    #And posterior covariance 'function'
    postcov <- Kyy - Kyx%*%Ainv%*%Kxy #calculate posterior covariance
  
    GP <- mvrnorm(1,postmean, postcov) #sample from GP
    delta <- exp(GP) #undo log transformation
  }



  #######
  #DetGP Prediction
  #######
  detGP <- function(inputx, inputy, predx, alpha, beta, l, nugget) {
    #generates a gaussian process for given inputs, outputs, prediction points, and parameter estimates
    #This GP will be for the deterministic data (which makes it a standard GP prediction)
  
    #We need A = K(X, X); Kyx = K(X*, X); Kyy = K(X*, X*); Kxy = K(X, X*)
    
    A <- (alpha^2) * K12(inputx, inputx, l) + diag(rep(nugget, length(inputx[,1])), nrow = length(inputx[,1])) #plus nugget 
    Ainv <- chol2inv(chol(A)) #get inverse
  
    Kyx <- (alpha^2) * K12(predx, inputx, l)
    Kxy <- t(Kyx)
    Kyy <- (alpha^2) * K12(predx, predx, l) + diag(rep(nugget, length(predx[,1])), nrow = length(predx[,1]))#plus nugget
  
    #Now get posterior mean 'function'
    #obtain prior mean functions for input and prediction points
    Hpred <- detH(predx)
    Hinput <- detH(inputx)
    postmean <- meanfunc(Hpred, beta) + Kyx%*%Ainv%*%(inputy- meanfunc(Hinput, beta)) #calculate posterior mean
  
    #And posterior covariance 'function'
    postcov <- Kyy - Kyx%*%Ainv%*%Kxy #calculate posterior covariance
  
    GP <- mvrnorm(1,postmean, postcov) #sample from GP
  }

  

  #######
  #HetGP Prediction
  #######
  HetGP <- function(inputx, inputy, predx, alpha, beta, l, inputdelta, preddelta, nugget, inputdet) {
    #generates a gaussian process for given inputs, outputs, prediction points, and parameter estimates
    #This GP will be for the actual output y (this means its a standard GP prediction, but with input dependent nugget, and the prior mean includes the detGP)
  
    #We need A = K(X, X); Kyx = K(X*, X); Kyy = K(X*, X*); Kxy = K(X, X*)
    
    A <- (alpha^2) * K12(inputx, inputx, l) + diag(rep(nugget, length(inputx[,1])), nrow = length(inputx[,1])) + diag(inputdelta, nrow = length(inputx[,1])) #plus nugget and delta
    Ainv <- chol2inv(chol(A))
  
    Kyx <- (alpha^2) * K12(predx, inputx, l)
    Kxy <- t(Kyx)
    Kyy <- (alpha^2) * K12(predx, predx, l) + diag(rep(nugget, length(predx[,1])), nrow = length(predx[,1])) + diag(preddelta, nrow = length(predx[,1])) #plus nugget and delta
  
    #Now get posterior mean 'function'
    #obtain prior mean functions for input and prediction points
    Hpred <- H(predx)
    Hinput <- H(inputx)
    postmean <- meanfunc(Hpred, beta) + Kyx%*%Ainv%*%(inputy - meanfunc(Hinput, beta) - inputdet) #calculate posterior mean
  
    #And posterior covariance 'function'
    postcov <- Kyy - Kyx%*%Ainv%*%Kxy #calculate posterior covariance
  
    GP <- mvrnorm(1,postmean, postcov) #sample from GP
  }



  #With these functions all set up, we can easily do the sampling.

  #let us sample the GPs M times. To make things easier to do this again for different prediction points, this is functionised
  GPfuncs <- function(inputx, inputy, predx, inputdetx, inputdety, parameterestimates, nugget, M){
    
    #initialise everything
    N <- length(inputx[,1]) #number of input points
    Npred <- length(predx[,1]) #number of prediction points
    preddeltaGPsamples <- matrix(0, ncol = Npred, nrow = M) #variance GP predictions
    detGPsamples <- matrix(0, ncol = N+Npred, nrow = M)     #DetGP predictions for the stochastic input data points and prediction points
    inputDetGPsamples <- matrix(0, ncol = N, nrow = M)      #DetGP predictions just for stochastic input points
    predDetGPsamples <- matrix(0, ncol = Npred, nrow = M)   #DetGP predictions just for prediction points
    HetGPsamples <- matrix(0, ncol = Npred, nrow = M)       #HetGP predictions
  
  
    for (i in 1:(M)) {
      #iteratively sample from the variance GP, detGP, and then HetGP
      
      #variance GP
      preddeltaGPsamples[i,] <- deltaGP(inputx, as.vector(parameterestimates$logdelta), predx, parameterestimates$alpha_var, as.vector(parameterestimates$beta_var), as.vector(parameterestimates$length_scale_var), nugget, parameterestimates$sigma)
      
      #DetGP
      detGPsamples[i,] <- detGP(inputdetx, inputdety, rbind(inputx, predx), parameterestimates$alpha_det, as.vector(parameterestimates$beta_det), as.vector(parameterestimates$length_scale_det), nugget)
      inputDetGPsamples[i,] <- detGPsamples[i, 1:N]
      predDetGPsamples[i,] <- detGPsamples[i, (N+1):(N+Npred)]
      
      #HetGP
      HetGPsamples[i,] <- HetGP(inputx, inputy, predx, parameterestimates$alpha, as.vector(parameterestimates$beta), as.vector(parameterestimates$length_scale), as.vector(parameterestimates$delta), preddeltaGPsamples[i,], nugget, inputDetGPsamples[i,])
    }
  
    #we then return the predictions we want (DetHetGP, variance GP, DetGP for inputs, DetGP for predictions, HetGP)
    list(DetHetGP = HetGPsamples + predDetGPsamples, preddeltaGPsamples, inputDetGPsamples, predDetGPsamples, HetGPsamples)
  }

  
  
  #number of samples to take
  M <- M
  nugget <- 0.0001 #nugget
  
  #then actually sample 
  system.time(GPfuncs <- GPfuncs(x, y, xpred, xdet, ydet, parameterestimates, nugget, M))

  GPfuncs #return it
}


#here is an example line(s) of code running this function
#samples <- DetHetGPPredict(x, y, xdet, ydet, parameterestimates, xpred)
#ysamples <- samples[[1]] #extract the DetHetGP predictions



