//This is the Stan code for the DetHetGP model. Essentially this is just the Bayesian Hierarchical model equations, 
//but with some computational tricks needed for Stan


//first we have to declare our data
data {
  //stochastic Data
  int<lower=1> N;         //number of stochastic observations
  int<lower=1> d;         //number of dimensions
  vector[N] y;            //function values at the observations
  row_vector[d] x[N];     //input values at the observations
  real<lower = 0> nugget; //nugget
  int<lower=1> dH;        //size of  GP H matrix
  matrix[N, dH] H;        //GP regression matrix
  int<lower=1> dHvar;     //size of variance GP H matrix
  matrix[N, dHvar] Hvar;  //variance GP regression matrix
  
  //Deterministic Data
  int<lower=1> Ndet;        //number of observations for the deterministic run
  row_vector[d] xdet[Ndet]; //input values at the deterministic observations
  vector[Ndet] ydet;        //function values at the deterministic observations
  int<lower=1> dHdet;       //size of deterministic GP H matrix
  matrix[Ndet, dHdet] Hdet; //det GP regression matrix
  matrix[N, dHdet] Hdet2;   //det GP regression matrix for stochastic points
  
}


//At the time of writing this, isotropic covariance functions weren't implemented, so this is a trick that allows them
//we rescale the data instead, and then input into covariance functions just an ``identity'' length scale.
//these lines construct said ``identity'' length scales
transformed data{
  real fake_length_scale;     //for the HetGP
  real fake_length_scale_var; //for the variance GP
  real fake_length_scale_det; //for the DetGP
  
  fake_length_scale = pow(sqrt(2), -1 );
  fake_length_scale_var = pow(sqrt(2), -1 );
  fake_length_scale_det = pow(sqrt(2), -1 );
}



//then we have to declare the parameters (we want to estimate)
parameters {
  //For the HetGP 
  row_vector<lower=0>[d] length_scale; 
  real<lower=0> alpha;
  vector[dH] beta;
  
  //for variance GP
  row_vector<lower=0>[d] length_scale_var;
  real<lower=0> alpha_var;
  vector[dHvar] beta_var;
  real<lower=0> sigma;
  
  vector[N] logdelta;
  
  
  //for det GP
  row_vector<lower=0>[d] length_scale_det_raw; //"raw"" because need to add small number to it as well, see appendix of paper
  real<lower=0> alpha_det;
  vector[dHdet] beta_det;
}


//Then we can "transform" said parameters. For fitting a GP, this involves constructing the mean and covariance functions
transformed parameters {
  
  //again we have to declare these first
  
  //for HetGP
  row_vector[d] xl[N];  //The transformed data, for the isotropic trick
  matrix[N,N] K;        //The covariance function at the inputs
  vector[N] mu;         //The mean function at the inputs
  
  //for variance GP
  row_vector[d] xlvar[N];     //The transformed data, for the isotropic trick
  matrix[N, N] Kvar;          //The covariance function at the inputs
  matrix[N, N] cholKvar;      //The cholesky decomposition of the covariance function
  vector[N] muvar;            //The mean function at the inputs
  vector<lower=0>[N] delta;   //The values of the variance at the inputs (to be estimated)
  
  //for det GP model (this ends up a little confusing, as two types of data end up getting refered to
  //anthing with a "det" suffix refers to the deterministic data, no suffix means its the stochastic data)
  row_vector<lower=0>[d] length_scale_det;  //The length scale (post adding the small number, see appendix of paper)
  row_vector[d] xldet[Ndet];                //The transformed data, for the isotropic trick
  matrix[Ndet, Ndet] Kdet;                  //The covariance function at the inputs
  matrix[Ndet, Ndet] cholKdet;              //The cholesky decomposition of the covariance function
  vector[Ndet] mudet;                       //The mean function at the inputs
  vector[N] mudet2;                         //The mean function at the (stochastic) inputs
  
  //Fitting DetHetGP also involves obtaining estimates for the DetGP at the stochastic inputs
  //These next objects are necessary to make said predictions from DetGP
  vector[Ndet] Kdet_div_y;                  
  matrix[Ndet, N] v_pred;
  matrix[N, Ndet] detKyx;
  matrix[N, N] detKyy;
  row_vector[d] xldetpred[N]; 
  vector[N] mudetpred;
  matrix[N,N] covdetpred;
  
  
  //final mean and covariances
  vector[N] ymu;      //The mean function at the inputs
  matrix[N, N] ycov;  //The covariance function at the inputs
  
  
  //now do fitting (this will involve referring to all the objects we've now assigned)
  
  
  //////////
  //variance GP
  /////////
  delta = exp(logdelta); //need to undo the log transform for the variances
  
  //do trick to allow non-isotropy
  for (i in 1:N)
    xlvar[i] = x[i]./length_scale_var; //transform the data
    
  //calculate mean func
  muvar = Hvar * beta_var;
  
  //build covariance function
  Kvar = cov_exp_quad(xlvar, alpha_var, fake_length_scale_var); 
  for (n in 1:N)
    Kvar[n, n] = Kvar[n, n] + nugget + sigma; // add nugget to diagonal elements
  
  cholKvar = cholesky_decompose(Kvar); //obtain cholesky decomposition
  
  
  
  //////////
  //Het GP
  /////////
  
  //do trick to allow non-isotropy
  for (i in 1:N)
    xl[i] = x[i]./length_scale; //transform the data
  
  //calculate mean func
  mu = H * beta;
  
  //build covariance function
  K = cov_exp_quad(xl, alpha, fake_length_scale);
  // diagonal elements
  for (n in 1:N)
    K[n, n] = K[n, n] + nugget + delta[n]; // add nugget (and intrinsic variance) to diagonal elements
  
  
  
  //////////
  //Det GP
  /////////
  
  length_scale_det = length_scale_det_raw + 0.05; //add small number to det length_scales (see appendix of paper)
  
  //do trick to allow non-isotropy
  for (i in 1:Ndet)
    xldet[i] = xdet[i]./length_scale_det; //transform the data

  //calculate mean func
  mudet = Hdet * beta_det;
  
  //mean func for stochastic data points
  mudet2 = Hdet2 * beta_det;
  
  //build covariance function
  Kdet = cov_exp_quad(xldet, alpha_det, fake_length_scale_det); 
  // diagonal elements
  for (n in 1:Ndet)
    Kdet[n, n] = Kdet[n, n] + nugget; // add nugget to diagonal elements
  
  cholKdet = cholesky_decompose(Kdet); //obtain cholesky decomposition
    
    
    
  //The next lines of code appear a bit ugly. They are just the GP predictive equations;
  //predicting the values of DetGP but for the stochastic points
  
  for (i in 1:N)
    xldetpred[i] = x[i]./length_scale_det;   //do trick to allow non-isotropy
    
  Kdet_div_y = mdivide_left_tri_low(cholKdet, ydet-mudet); //computational tricks to speed up next few lines
  Kdet_div_y = mdivide_right_tri_low(Kdet_div_y',cholKdet)';
    
  detKyx = cov_exp_quad(xldetpred, xldet, alpha_det, fake_length_scale_det);  //obtain K(X*, X)
  detKyy = cov_exp_quad(xldetpred, alpha_det, fake_length_scale_det);         //obtain K(X*, X*)
  
  for (n in 1:N)
    detKyy[n, n] = detKyy[n, n] + nugget; //add nugget to K(X*, X*)
  
  mudetpred = mudet2 + detKyx*Kdet_div_y;           //obtain predictive values for the mean of DetGP (at stochastic data points)
  v_pred = mdivide_left_tri_low(cholKdet, detKyx'); //computational trick to speed up next line
  covdetpred = detKyy - v_pred'*v_pred;             //obtain predictive values for the covariance of DetGP (at stochastic data points)
  
  
  
  
  //////////
  //final mean and covariances (obtained by summing HetGP with DetGP)
  //////////
  ymu = mudetpred + mu; //final mean = mean from DetGP + mean from HetGP
  ycov = covdetpred + K; //final covariance = covariance from DetGP + covariance from HetGP
  
}



//now we specifiy the priors (for the hyperparameters (which one can change), and the GP "priors")
model {

//priors for HetGP
alpha ~ inv_gamma(2,1);
length_scale ~ gamma(4, 4);
beta ~ normal(0, 10);

//priors for variance GP
alpha_var ~ inv_gamma(2,1);
length_scale_var ~ gamma(4, 4);
beta_var ~ normal(0, 10);
sigma ~ normal(0,1);

//priors for det GP
alpha_det ~ inv_gamma(2,1);
beta_det ~ normal(0, 10);
length_scale_det_raw ~ gamma(4, 4);



//And now, the GP "priors" 

//DetGP
ydet ~ multi_normal(mudet, Kdet);

//(log) variance GP
logdelta ~ multi_normal_cholesky(muvar, cholKvar);

//DetHetGP 
//Its not a "HetGP prior" because: y-detGP ~ HetGP => y ~ HetGP+DetGP => y ~ N(muHet+muDet, CovHet+CovDet)
y ~ multi_normal(ymu, ycov);

}



