//This is the Stan code for the HetGP model. Essentially this is just the Bayesian Hierarchical model equations, 
//but with some computational tricks needed for Stan

//first we have to declare our data
data {
  //stochastic Data
  int<lower=1> N;         //number of stochastic observations
  int<lower=1> d;         //number of dimensions
  vector[N] y;            //function values at the observations
  row_vector[d] x[N];     //input values at the observations
  real<lower = 0> nugget; //nugget
  int<lower=1> dH;        //size of  GP H matrix
  matrix[N, dH] H;        //GP regression matrix
  int<lower=1> dHvar;     //size of variance GP H matrix
  matrix[N, dHvar] Hvar;  //variance GP regression matrix
}


//At the time of writing this, isotropic covariance functions weren't implemented, so this is a trick that allows them
//we rescale the data instead, and then input into covariance functions just an ``identity'' length scale.
//these lines construct said ``identity'' length scales
transformed data{
  real fake_length_scale;     //for the HetGP
  real fake_length_scale_var; //for the variance GP

  fake_length_scale = pow(sqrt(2), -1 );
  fake_length_scale_var = pow(sqrt(2), -1 );
}




//then we have to declare the parameters (we want to estimate)
parameters {
  //For the HetGP 
  row_vector<lower=0>[d] length_scale; 
  real<lower=0> alpha;
  vector[dH] beta;
  
  //for variance GP
  row_vector<lower=0>[d] length_scale_var;
  real<lower=0> alpha_var;
  vector[dHvar] beta_var;
  real<lower=0> sigma;
  
  vector[N] logdelta;
}


//Then we can "transform" said parameters. For fitting a GP, this involves constructing the mean and covariance functions
transformed parameters {
  
  //again we have to declare these first
  
  //for HetGP
  row_vector[d] xl[N];  //The transformed data, for the isotropic trick
  matrix[N,N] K;        //The covariance function at the inputs
  matrix[N, N] cholK;   //The cholesky decomposition of the covariance function
  vector[N] mu;         //The mean function at the inputs
  
  //for variance GP
  row_vector[d] xlvar[N];     //The transformed data, for the isotropic trick
  matrix[N, N] Kvar;          //The covariance function at the inputs
  matrix[N, N] cholKvar;      //The cholesky decomposition of the covariance function
  vector[N] muvar;            //The mean function at the inputs
  vector<lower=0>[N] delta;   //The values of the variance at the inputs (to be estimated)
  
  //now do fitting (this will involve referring to all the objects we've now assigned)
  
  
  //////////
  //variance GP
  /////////
  delta = exp(logdelta); //need to undo the log transform for the variances
  
  //do trick to allow non-isotropy
  for (i in 1:N)
    xlvar[i] = x[i]./length_scale_var; //transform the data
    
  //calculate mean func
  muvar = Hvar * beta_var;
  
  //build covariance function
  Kvar = cov_exp_quad(xlvar, alpha_var, fake_length_scale_var); 
  for (n in 1:N)
    Kvar[n, n] = Kvar[n, n] + nugget + sigma; // add nugget to diagonal elements
  
  cholKvar = cholesky_decompose(Kvar); //obtain cholesky decomposition
  
  
  
  //////////
  //Het GP
  /////////
  
  //do trick to allow non-isotropy
  for (i in 1:N)
    xl[i] = x[i]./length_scale; //transform the data
  
  //calculate mean func
  mu = H * beta;
  
  //build covariance function
  K = cov_exp_quad(xl, alpha, fake_length_scale);
  // diagonal elements
  for (n in 1:N)
    K[n, n] = K[n, n] + nugget + delta[n]; // add nugget (and intrinsic variance) to diagonal elements
  
  cholK = cholesky_decompose(K); //obtain cholesky decomposition
  
}



//now we specifiy the priors (for the hyperparameters (which one can change), and the GP "priors")
model {
  
  //priors for HetGP
  alpha ~ inv_gamma(2,1);
  length_scale ~ gamma(4, 4);
  beta ~ normal(0, 10);

  //priors for variance GP
  alpha_var ~ inv_gamma(2,1);
  length_scale_var ~ gamma(4, 4);
  beta_var ~ normal(0, 10);
  sigma ~ normal(0,1);


  
  //And now, the GP "priors" 

  //(log) variance GP
  logdelta ~ multi_normal_cholesky(muvar, cholKvar);
  
  //HetGP
  y ~ multi_normal_cholesky(mu, cholK);
}



